@@ -19,7 +19,7 @@ use Win32::File::Object;
 use IPC::Run3;
 use Digest::SHA1;
 
-##### manatory methods for all Step-like classess - new(), check(), run(), test()
+##### mandatory methods for all Step-like classes - new(), check(), run(), test()
 
 sub new {
   my $class = shift;
@@ -220,10 +220,6 @@ sub _convert_name {
   return $local_name;
 }
 
-sub _insert_fragment {
-#XXX-FIXME remove this method
-}
-
 sub get_path_string {
   my $self = shift;
 
@@ -362,23 +358,99 @@ sub sha1_file {
   return $sha1->hexdigest;
 }
 
-sub workaround_get_dist_list {
-# XXX-FIXME ugly workaround for getting the list of distributions installed  via CPANPLUS
-# used by UpgradeCpanModules and InstallModules
-# implement something more clever in the future
-  my $self = shift;
-  my $env = { PERL5_CPANPLUS_HOME=>$self->global->{build_ENV}->{APPDATA} }; #workaround for CPANPLUS
-  my $script_pl = $self->boss->resolve_name("<dist_sharedir>/utils/CPANPLUS_list_build_cache.pl");
-  my $prefix = "cpan_dist_list.".time;
-  my $log = catfile($self->global->{debug_dir}, "$prefix.log.txt");
-  my $dumper = catfile($self->global->{debug_dir}, "$prefix.dumper.txt");
-  my $nstore = catfile($self->global->{debug_dir}, "$prefix.nstore.txt");
-  my $rv = $self->execute_special(['perl', $script_pl, '-out_nstore', $nstore, '-out_dumper', $dumper ], $log, $log, $env);
-  die "ERROR: exec '$script_pl' failed" unless defined $rv && $rv == 0;
-  die "ERROR: missing file '$nstore'" unless -f $nstore;
-  my $data = retrieve($nstore) or die "ERROR: retrieve failed, probably error while executing '$script_pl'";
-  die "ERROR: invalid data" unless defined $data && ref $data eq 'ARRAY';
-  return $data;
+sub install_modlist {
+  my ($self, @list) = @_;
+
+  return 1 unless @list;
+  my $success = 1;
+  my @distlist_final = ();
+
+  $self->boss->message(1, "NOTE: global option test_modules=0 (all tests will be skipped") unless $self->global->{test_modules};
+
+  my ($distlist, $rv);
+  my $count = scalar(@list);
+  my $i = 0;
+  for my $item (@list) {
+    $i++;
+    $item = { module=>$item } unless ref $item; # if item is scalar we assume module name
+    my $name = $item->{module};
+    if ($name) {
+      my @msg;
+      push @msg, 'IGNORE_TESTFAILURE' if $item->{ignore_testfailure};
+      push @msg, 'SKIPTEST' if $item->{skiptest};
+      $self->boss->message(1, sprintf("installing %2d/%d '%s' \t".join(' ',@msg), $i, $count, $name));
+      ($distlist, $rv) = $self->_install_module(%$item);
+      push @distlist_final, @$distlist;
+      unless(defined $rv && $rv == 0) {
+        $self->boss->message(1, "WARNING: non-zero exit code '$rv' - gonna continue but overall result of this task will be 'FAILED'");
+        $success = 0;
+      }
+    }
+    else {
+      $self->boss->message(1, sprintf("SKIPPING!! %2d/%d ERROR: invalid item", $i, $count));
+      $success = 0;
+    }
+  }
+  
+  $self->boss->message(2, "WARNING: empty distribution_list (that's not good)") unless scalar(@distlist_final)>0;
+  
+  # store some output data
+  $self->{data}->{output}->{distributions} = \@distlist_final;
+
+  return $success;
+}
+
+sub _install_module {
+  my ($self, %args) = @_;
+
+  my $now = time;
+
+  my $shortname = $args{module};
+  $shortname =~ s|^.*[\\/]||;
+  $shortname =~ s|:+|_|g;
+  $shortname =~ s|[\\/]+|_|g;
+  $shortname =~ s/\.(tar\.gz|tar\.bz2|zip|tar|gz)$//;
+  my $script_pl = $self->boss->resolve_name("<dist_sharedir>/utils/CPANMINUS_install_module.pl");
+  my $log         = catfile($self->global->{debug_dir}, "mod_install_".$now."_".$shortname.".log.txt");
+  my $dumper_file = catfile($self->global->{debug_dir}, "mod_install_".$now."_".$shortname.".list.dumper.txt");
+  my $nstore_file = catfile($self->global->{debug_dir}, "mod_install_".$now."_".$shortname.".list.nstore.txt");
+
+  my $env = {
+    PERL_MM_USE_DEFAULT=>1, AUTOMATED_TESTING=>undef, RELEASE_TESTING=>undef,
+    PERL5_CPANPLUS_HOME=>$self->global->{build_ENV}->{APPDATA}, #workaround for CPANPLUS
+  };
+  # resolve macros in env{}
+  if (defined $args{env} && ref $args{env} eq 'HASH') {
+    for my $var (keys %{$args{env}}) { 
+      $env->{$var} = $self->boss->resolve_name($args{env}->{$var});
+    }
+  }
+  # resolve macros in module name
+  $args{module} = $self->boss->resolve_name($args{module});
+  $args{module} =~ s|\\|/|g; # cpanm dislikes backslashes
+
+  my %params = ( '-url' => $self->global->{cpan_url}, '-install_to' => 'vendor', '-module' => $args{module} ); #XXX-TODO multiple modules?
+  $params{'-out_dumper'}         = $dumper_file if $dumper_file;
+  $params{'-out_nstore'}         = $nstore_file if $nstore_file;
+  $params{'-install_to'}         = $args{install_to}         if defined $args{install_to};
+  $params{'-verbose'}            = $args{verbose}            if defined $args{verbose};
+  $params{'-skiptest'}           = $args{skiptest}           if defined $args{skiptest};
+  $params{'-ignore_testfailure'} = $args{ignore_testfailure} if defined $args{ignore_testfailure};
+  $params{'-ignore_uptodate'}    = $args{ignore_uptodate}    if defined $args{ignore_uptodate};
+  $params{'-prereqs'}            = $args{prereqs}            if defined $args{prereqs};
+  $params{'-interactivity'}      = $args{interactivity}      if defined $args{interactivity};
+  $params{'-makefilepl_param'}   = $args{makefilepl_param}   if defined $args{makefilepl_param}; #XXX-TODO multiple args?
+  $params{'-buildpl_param'}      = $args{buildpl_param}      if defined $args{buildpl_param};    #XXX-TODO multiple args?
+  # handle global test skip
+  $params{'-skiptest'} = 1 unless $self->global->{test_modules};
+  # Execute the module install script
+  my $rv = $self->execute_special(['perl', $script_pl, %params], $log, $log, $env);
+  unless(defined $rv && $rv == 0) {
+    rename $log, catfile($self->global->{debug_dir}, "mod_install_FAIL_".$now."_".$shortname.".log.txt");
+    return [], $rv;
+  }
+  my $data = retrieve($nstore_file) or die "ERROR: retrieve failed";
+  return ($data->{installed}//[]), $rv;
 }
 
 1;
\ No newline at end of file
